import argparse
from pathlib import Path
import subprocess
import sys
from typing import List

blade_managed_source_folders = [
    'plugins/PPBlade',
    'plugins/MeshTools',
    'plugins/OrientationPlugin',
    'plugins/PrimeTowerConfigurator',
]
ignored_files = [
    "**/__init__*"
]

def lint(file: Path) -> bool:
    print(f"Applying linting to file: {file}")
    command = ["pylint", "--rcfile=.pylintrc", str(file)]
    result = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True, check=False)

    if result.returncode != 0:
        print(file)
        print("Nonzero exit status: ", result.returncode, file=sys.stderr)
        if result.stderr:
            print("Error messages: ", result.stderr, file=sys.stderr)
        print(result.stdout, file=sys.stderr)

        return False

    return True

def format_pep8(file: Path) -> None:
    print(f"Applying autopep8 to file: {file}")
    command = ["python","-m","autopep8", "--max-line-length", "150", "-i", str(file)]
    result = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True, check=True)
    print(result.stdout, result.stderr)

def format_mypy(dir:Path) -> bool:
    print(f"Applying mypy to dir: {dir}")
    args = ['mypy','.']
    result = subprocess.run(args,cwd=dir,stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    if result!=0:
        print("Nonzero exit status: ", result.returncode, file=sys.stderr)
        if result.stderr:
            print("Error messages: ", result.stderr, file=sys.stderr)
        print(result.stdout, file=sys.stderr)
    
def should_not_ignore(file: Path, ignore_list: List[str]) -> bool:
    for ignored in ignore_list:
        for single_ignore in Path().glob(ignored):
            if file==single_ignore:
                return False

    return True


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog='pylinter',
        description=f'Performs python linting on the client source code',
        epilog='Text at the bottom of help')
    parser.add_argument('-f', '--format',action='store_true',help='A flag to apply formatting to the blade managed python source code')  # on/off flag
    parser.add_argument('-l', '--lint',action='store_true',help='A flag to apply linting')  # on/off flag
    parser.add_argument('-r', '--report',action='store_true',help='TODO: A flag to save out a report')  # on/off flag
    parser.add_argument('-m','--mypy',action='store_true',help='Perform mypy linting')
    args = parser.parse_args()

    if args.format:
        for directory in blade_managed_source_folders:
            p = Path(directory)
            for python_file in p.glob('**/*.py'):
                format_pep8(python_file)
    
    if args.lint:
        res = True
        for directory in blade_managed_source_folders:
            p = Path(directory)
            for python_file in p.glob('**/*.py'):
                if should_not_ignore(python_file,ignored_files):
                    res = res & lint(python_file)

        # Don't report for now. Should be enabled!
        if not res:
            sys.exit(666)
    
    if args.mypy:
        res = True
        for directory in blade_managed_source_folders:
            p = Path(directory)
            res = res & format_mypy(p)
        
        if not res:
            sys.exit(666)