// Copyright (c) 2022 Ultimaker B.V.
// Cura is released under the terms of the LGPLv3 or higher.
// Modified by BigRep GmbH 2024

import QtQuick 2.7
import QtQuick.Layouts 1.1
import QtQuick.Controls 2.3

import UM 1.5 as UM
import Cura 1.1 as Cura

Item
{
    id: machineSelector

    property alias buttonsMajor: machineSelectorMajor.buttons
    property alias machineManagerMajor: machineSelectorMajor.machineManager
    property alias headerCornerSideMajor: machineSelectorMajor.headerCornerSide
    function toggleContentMajor() {
        machineSelectorMajor.toggleContent() 
    }

    property alias buttonsMinor: machineSelectorMinor.buttons
    property alias machineManagerMinor: machineSelectorMinor.machineManager
    property alias machineListModelMinor: machineSelectorListMinor.model
    property alias onSelectPrinterMinor: machineSelectorListMinor.onSelectPrinter
    property alias headerCornerSideMinor: machineSelectorMinor.headerCornerSide
    function toggleContentMinor() {
        machineSelectorMinor.toggleContent() 
    }

    property var machineType

    RowLayout
    {
        width: parent.width
        height: parent.height
        spacing: 0 

        Cura.ExpandablePopup
        {
            id: machineSelectorMajor

            width: parent.width / 2
            Layout.fillWidth: true
            Layout.fillHeight: true

            property Cura.MachineManager machineManager
            property bool isNetworkPrinter: machineManager.activeMachineHasNetworkConnection
            property bool isConnectedCloudPrinter: machineManager.activeMachineHasCloudConnection
            property bool isCloudRegistered: machineManager.activeMachineHasCloudRegistration
            property bool isGroup: machineManager.activeMachineIsGroup
            property string machineName: machineManager.activeMachine.definition.name

            property list<Item> buttons

            contentPadding: UM.Theme.getSize("default_lining").width
            contentAlignment: Cura.ExpandablePopup.ContentAlignment.AlignLeft

            UM.I18nCatalog
            {
                id: catalogMajor
                name: "cura"
            }

            headerItem: Cura.IconWithText
            {
                text: machineSelectorMajor.machineName
                property string iconName: machineSelectorMajor.machineManager.activeMachine.getIcon()
                source: iconName ? UM.Theme.getIcon(iconName) : UM.Theme.getIcon("PrinterOne", "medium")
                font: UM.Theme.getFont("medium")
                iconColor: UM.Theme.getColor("machine_selector_printer_icon")
                iconSize: source != "" ? UM.Theme.getSize("machine_selector_icon").width: 0

                Component.onCompleted: {
                    console.log("iconName")
                    console.warn(iconName)
                }
            }

            property int minDropDownWidth: UM.Theme.getSize("machine_selector_widget_content").width
            property int maxDropDownHeight: UM.Theme.getSize("machine_selector_widget_content").height

            contentItem: Item
            {
                id: popupMajor
                implicitWidth: Math.max(machineSelectorMajor.width, machineSelectorMajor.minDropDownWidth)
                implicitHeight: Math.min(machineSelectorListMajor.contentHeight + separatorMajor.height + buttonRowMajor.height, machineSelectorMajor.maxDropDownHeight) //Maximum height is the theme entry.

                Cura.MachineTypeListModel
                {
                    id: machineTypeModelMajor
                }

                MachineTypeSelectorList
                {
                    model: machineTypeModelMajor
                    
                    id: machineSelectorListMajor
                    anchors
                    {
                        left: parent.left
                        leftMargin: UM.Theme.getSize("default_lining").width
                        right: parent.right
                        rightMargin: UM.Theme.getSize("default_lining").width
                        top: parent.top
                        bottom: separatorMajor.top
                    }
                    clip: true
                }

                Rectangle
                {
                    id: separatorMajor
                    anchors.bottom: buttonRowMajor.top
                    width: parent.width
                    height: UM.Theme.getSize("default_lining").height
                    color: UM.Theme.getColor("lining")
                }

                Row
                {
                    id: buttonRowMajor

                    anchors.bottom: parent.bottom
                    anchors.left: parent.left
                    anchors.right: parent.right

                    padding: UM.Theme.getSize("default_margin").width
                    spacing: UM.Theme.getSize("default_margin").width

                    children: buttonsMajor
                }

                states: [
                    State {
                        name: "noButtons"
                        when: !buttonsMajor || buttonsMajor.length == 0
                        PropertyChanges
                        {
                            target: buttonRowMajor
                            height: 0
                            padding: 0
                        }
                        PropertyChanges
                        {
                            target: separatorMajor
                            height: 0
                        }
                    }
                ]
            }
        }

        Cura.ExpandablePopup
        {
            id: machineSelectorMinor

            width: parent.width / 2
            Layout.fillWidth: true
            Layout.fillHeight: true

            property Cura.MachineManager machineManager
            property bool isNetworkPrinter: machineManager.activeMachineHasNetworkConnection
            property bool isConnectedCloudPrinter: machineManager.activeMachineHasCloudConnection
            property bool isCloudRegistered: machineManager.activeMachineHasCloudRegistration
            property bool isGroup: machineManager.activeMachineIsGroup
            property string machineName: {
                if (machineSelectorMinor.isNetworkPrinter && machineManager.activeMachineNetworkGroupName != "")
                {
                    return machineManager.activeMachineNetworkGroupName
                }
                if (machineSelectorMinor.isConnectedCloudPrinter)
                {
                    return machineManager.activeMachine.getName()
                }
                if (machineManager.activeMachine != null)
                {
                    return machineManager.activeMachine.getName()
                }
                return ""
            }


            property list<Item> buttons

            property string connectionStatus: {
                if (machineSelectorMinor.isNetworkPrinter)
                {
                    return "printer_connected"
                }
                else if (machineSelectorMinor.isConnectedCloudPrinter && Cura.API.connectionStatus.isInternetReachable)
                {
                    return "printer_cloud_connected"
                }
                else if (machineSelectorMinor.isCloudRegistered)
                {
                    return "printer_cloud_not_available"
                }
                else
                {
                    return ""
                }
            }

            function getConnectionStatusMessage() {
                if (connectionStatus == "printer_cloud_not_available")
                {
                    if(Cura.API.connectionStatus.isInternetReachable)
                    {
                        if (Cura.API.account.isLoggedIn)
                        {
                            if (machineManager.activeMachineIsLinkedToCurrentAccount)
                            {
                                return catalogMinor.i18nc("@status", "The cloud printer is offline. Please check if the printer is turned on and connected to the internet.")
                            }
                            else
                            {
                                return catalogMinor.i18nc("@status", "This printer is not linked to your account. Please visit the Ultimaker Digital Factory to establish a connection.")
                            }
                        }
                        else
                        {
                            return catalogMinor.i18nc("@status", "The cloud connection is currently unavailable. Please sign in to connect to the cloud printer.")
                        }
                    }
                    else
                    {
                        return catalogMinor.i18nc("@status", "The cloud connection is currently unavailable. Please check your internet connection.")
                    }
                }
                else
                {
                    return ""
                }
            }

            contentPadding: UM.Theme.getSize("default_lining").width
            contentAlignment: Cura.ExpandablePopup.ContentAlignment.AlignLeft

            UM.I18nCatalog
            {
                id: catalogMinor
                name: "cura"
            }

            headerItem: Cura.IconWithText
            {
                text: machineSelectorMinor.machineName
            }

            property int minDropDownWidth: UM.Theme.getSize("machine_selector_widget_content").width
            property int maxDropDownHeight: UM.Theme.getSize("machine_selector_widget_content").height

            contentItem: Item
            {
                id: popupMinor
                implicitWidth: Math.max(machineSelectorMinor.width, machineSelectorMinor.minDropDownWidth)
                implicitHeight: Math.min(machineSelectorListMinor.contentHeight + separatorMinor.height + buttonRowMinor.height, machineSelectorMinor.maxDropDownHeight) //Maximum height is the theme entry.
                MachineSelectorList
                {
                    id: machineSelectorListMinor
                    anchors
                    {
                        left: parent.left
                        leftMargin: UM.Theme.getSize("default_lining").width
                        right: parent.right
                        rightMargin: UM.Theme.getSize("default_lining").width
                        top: parent.top
                        bottom: separatorMinor.top
                    }
                    clip: true
                }

                Rectangle
                {
                    id: separatorMinor
                    anchors.bottom: buttonRowMinor.top
                    width: parent.width
                    height: UM.Theme.getSize("default_lining").height
                    color: UM.Theme.getColor("lining")
                }

                Row
                {
                    id: buttonRowMinor

                    anchors.bottom: parent.bottom
                    anchors.left: parent.left
                    anchors.right: parent.right

                    padding: UM.Theme.getSize("default_margin").width
                    spacing: UM.Theme.getSize("default_margin").width

                    children: buttonsMinor
                }

                states: [
                    State {
                        name: "noButtons"
                        when: !buttonsMinor || buttonsMinor.length == 0
                        PropertyChanges
                        {
                            target: buttonRowMinor
                            height: 0
                            padding: 0
                        }
                        PropertyChanges
                        {
                            target: separatorMinor
                            height: 0
                        }
                    }
                ]
            }
        }
    }
}